#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/param.h> // for MAXPATHLEN
#include <time.h>
#include <assert.h>

#include <fitsio.h>
#include <librdf.h>
#include <rdf_concepts.h>       /* supposed to be redland.h, but that doesn't work */

#include "config.h"
#include "hgstamp.h"

void fitserror(int status);
void librdferror(char*);
void general_error(char* fmt, ...);
double get_value_double(fitsfile* fptr, char* keyname);
int get_value_int(fitsfile* fptr, char* keyname);

int serialize_model(librdf_model*, librdf_uri*, char*, FILE*);

// Global constants
librdf_world* world;

#define FITS_CORE_BASE "http://purl.org/astro/fits/core#"
#define FITS_CTYPE_BASE "http://purl.org/astro/fits/ctype#"
#define DC_BASE "http://purl.org/dc/terms/"
#define FOAF_BASE "http://xmlns.org/foaf/1.1/"
#define PROVENANCE "http://purl.org/net/provenance/ns#"

// shortcuts of various types
librdf_node* fits_resource_node(char*);
librdf_uri* xsd_predicate_uri(char* fragment);
librdf_node* new_typed_node(char*,char*);
librdf_uri* new_typed_uri(char* base, char* fragment);
librdf_node* new_internal_node();
#define RDF_NODE(key) librdf_new_node_from_node(LIBRDF_RS_ ## key(world))
#define RDF_TYPE librdf_new_node_from_node(LIBRDF_MS_type(world))

// convenience functions
int add_array_predicate(fitsfile* fptr,
                        librdf_model* model,
                        librdf_node* wcs_node,
                        char* fits_prefix,
                        char* rdf_property,
                        int naxis);
//librdf_node* get_node_list(librdf_model*, librdf_node*** node_list_ptr, int n, char*);
librdf_node* get_new_list_node(librdf_model* model, librdf_node* previous, librdf_node* first);
int add_double_predicate(fitsfile*, librdf_model*,librdf_node*,char*,char*);

// functions which add significant information -- logic within them
int add_provenance_information(librdf_model* model, librdf_node* wcs_node);
int add_extra_header_information(fitsfile* fptr, librdf_model* model, librdf_node* header_node);
int add_pixel_axes(fitsfile* fptr, librdf_model* model, librdf_node* wcs_node, int naxis);
int add_world_axes(fitsfile* fptr, librdf_model* model, librdf_node* wcs_node, int naxis);

// miscellaneous utilities
char* allocate_at_least(char* p, int required_len, size_t *current_len);
char* format(int, char* fmt, ...);
#define UFORMAT (unsigned char*)format
unsigned char* darray_to_string(double*, int);

char* progname;
void Usage();

#define LIBRDF_ASSIGN(val, call, ...)                   \
    if ((val = librdf_ ## call(__VA_ARGS__)) == NULL) { \
        librdferror("Can't call " # call);              \
}
#define LIBRDF_CALL(call, ...)                   \
    if (librdf_ ## call(__VA_ARGS__)) { \
        librdferror("Can't call " # call);              \
}
#define MALLOC(var, type, size)                                         \
    if ((var = (type*)malloc((size) * sizeof(type))) == NULL) {         \
    fprintf(stderr, "Malloc failed at %s, line %d: unable to allocate %d " #type, __FILE__, __LINE__, (size)); \
    exit(1);                                                            \
    }
#define REALLOC(var, type, size)                                         \
    if ((var = (type*)realloc(var, (size) * sizeof(type))) == NULL) {   \
    fprintf(stderr, "Realloc failed at %s, line %d: unable to allocate %d " #type, __FILE__, __LINE__, (size)); \
    exit(1);                                                            \
    }
    

int main(int argc, char** argv)
{
    char* filename = NULL;
    unsigned char* base_uri_string = NULL;
    fitsfile *fptr;
    int status = 0;
    int naxis;
    char* serialization_format = "turtle";
    int ch;

    librdf_model* model;
    librdf_uri* base_uri;
    librdf_node* base_node;
    librdf_node* wcs_node;

    LIBRDF_ASSIGN(world, new_world);
    librdf_world_open(world);

    progname = argv[0];
    while ((ch = getopt(argc, argv, "s:")) != -1) {
        switch (ch) {
          case 's':
            if (librdf_serializer_check_name(world, optarg)) {
                serialization_format = optarg;
            } else {
                int i = 0;
                const char* serializer_name;
                fprintf(stderr, "Unknown serialization format: options are:");
                for (i=0; librdf_serializer_enumerate(world, i, &serializer_name, NULL) == 0; i++) {
                    fprintf(stderr, " %s", serializer_name);
                }
                fputc('\n', stderr);
                Usage();
            }
            break;
          default:
            Usage();
        }
    }
    argc -= optind;
    argv += optind;
    for (; argc>0; argc--, argv++) {
        if (filename == NULL) {
            filename = *argv;
        } else if (base_uri_string == NULL) {
            base_uri_string = (unsigned char*)*argv;
        } else {
            Usage();
        }
    }

    if (filename == NULL) {
        Usage();
    }
    if (base_uri_string == NULL) {
        if (filename[0] == '/') {
            base_uri_string = UFORMAT(1, "file:%s", filename);
        } else {
            char cwd[MAXPATHLEN];
            base_uri_string = UFORMAT(1, "file:%s/%s", getcwd(cwd, MAXPATHLEN), filename);
        }
        //fprintf(stderr, "base-uri set to %s\n", base_uri_string);
    }

    // All setup done -- get going
    
    status = 0;

    // Open the FITS file.  If there's no header number given in the
    // 'filename', then this will move to "the first HDU containing
    // significant data", as described in the cfitsio manual.
    if (fits_open_data(&fptr, filename, READONLY, &status)) {
        fitserror(status);
    }

    LIBRDF_ASSIGN(model, new_model, world, librdf_new_storage(world, NULL, NULL, NULL), NULL);

    LIBRDF_ASSIGN(base_uri, new_uri, world, base_uri_string);
    LIBRDF_ASSIGN(base_node, new_node_from_uri, world, base_uri);
    LIBRDF_CALL(model_add, model, base_node, RDF_TYPE, fits_resource_node("FITSheader"));
    wcs_node = new_internal_node();
    LIBRDF_CALL(model_add, model, base_node, fits_resource_node("hasWCSInformation"), wcs_node);
    LIBRDF_CALL(model_add, model, wcs_node, RDF_TYPE, fits_resource_node("WCSInformation"));

    // Now start extracting information from the FITS file and building up the model
    add_provenance_information(model, wcs_node);

    naxis = get_value_int(fptr, "naxis");
    if (naxis < 1) {
        fprintf(stderr, "That header has an unexpected NAXIS: %d\n", naxis);
        exit(1);
    }

    add_extra_header_information(fptr, model, base_node);

    add_pixel_axes(fptr, model, wcs_node, naxis);
    add_world_axes(fptr, model, wcs_node, naxis);
    
    add_array_predicate(fptr, model, wcs_node, "PC", "pcMatrix", naxis);
    add_array_predicate(fptr, model, wcs_node, "CD", "cdMatrix", naxis);

    // all done
    serialize_model(model, base_uri, serialization_format, stdout);

    fits_close_file(fptr, &status);
    librdf_free_world(world);
    exit(status);
}

int add_provenance_information(librdf_model* model, librdf_node* wcs_node)
{
    librdf_node* publisher = new_internal_node();
    librdf_node* publication = new_internal_node();
    struct tm* now;
    time_t now_time;
    char time_buffer[20];
        
    LIBRDF_CALL(model_add, model,
                wcs_node,
                new_typed_node(PROVENANCE, "retrievedBy"),
                publication);
    LIBRDF_CALL(model_add, model,
                publication,
                RDF_TYPE,
                new_typed_node(PROVENANCE, "DataAccess"));
    LIBRDF_CALL(model_add, model,
                publication,
                new_typed_node(PROVENANCE, "performedBy"),
                publisher);
    LIBRDF_CALL(model_add, model,
                publisher,
                RDF_TYPE,
                new_typed_node(PROVENANCE, "NonHumanActor"));
    LIBRDF_CALL(model_add_string_literal_statement,
                model,
                publisher, new_typed_node(FOAF_BASE, "name"),
                UFORMAT(0, "%s, %s", version_string, hg_version_stamp),
                NULL, 0);
    now_time = time(NULL);
    now = localtime(&now_time);
    strftime(time_buffer, 20, "%FT%T", now);
    LIBRDF_CALL(model_add_typed_literal_statement,
                model,
                publisher,
                new_typed_node(PROVENANCE, "performedAt"),
                (unsigned char*)time_buffer, NULL,
                xsd_predicate_uri("dateTime"));

    // Add here things like 'ORIGIN', FITSDATE, DATE, IRAF-TLM, etc...
    // (the provenance vocabulary has properties for these data creating things, too)

    return 0;
}

int add_extra_header_information(fitsfile* fptr, librdf_model* model, librdf_node* header_node)
{
    LIBRDF_CALL(model_add_typed_literal_statement,
                model, header_node, fits_resource_node("bitsPerPixel"),
                UFORMAT(0, "%d", get_value_int(fptr, "bitpix")), NULL,
                xsd_predicate_uri("integer"));
    return 0;
}


int add_pixel_axes(fitsfile* fptr, librdf_model* model, librdf_node* wcs_node, int naxis)
{
    librdf_node* axis_list_head = NULL;
    librdf_node* axis_list_p = NULL;
    int axisnumber;
    int status = 0;
    double dvalue;
    int ivalue;

    assert(naxis >= 1);

    for (axisnumber=1; axisnumber<=naxis; axisnumber++) {
        librdf_node* axis = librdf_new_node_from_uri_string(world, UFORMAT(0, "#pa%d", axisnumber));
        librdf_node* previous = axis_list_p;

        axis_list_p = get_new_list_node(model, previous, axis);
        if (axis_list_head == NULL) {
            axis_list_head = axis_list_p;
        }

        librdf_model_add(model,
                         axis,
                         RDF_TYPE,
                         fits_resource_node("PixelAxis"));
        librdf_model_add_typed_literal_statement(model,
                                                 axis,
                                                 fits_resource_node("axisNumber"),
                                                 UFORMAT(0, "%d", axisnumber), NULL,
                                                 xsd_predicate_uri("integer"));
        // CRPIXi
        if (fits_read_key(fptr, TDOUBLE, format(0, "CRPIX%d", axisnumber), &dvalue, NULL, &status)) {
            dvalue = 0.0;   /* default for CRPIXi */
            status = 0;
        }
        librdf_model_add_typed_literal_statement(model,
                                                 axis,
                                                 fits_resource_node("referencePixel"),
                                                 UFORMAT(0, "%g", dvalue), NULL,
                                                 xsd_predicate_uri("double"));

        // NAXISj
        if (fits_read_key(fptr, TINT, format(0, "NAXIS%d", axisnumber), &ivalue, NULL, &status)) {
            // this is an error
            general_error("No value for NAXIS%d", axisnumber);
        }
        librdf_model_add_typed_literal_statement(model,
                                                 axis,
                                                 fits_resource_node("numberOfPixels"),
                                                 UFORMAT(0, "%d", ivalue), NULL,
                                                 xsd_predicate_uri("integer"));

        // Link this axis to the WCS information
        librdf_model_add(model,
                         wcs_node,
                         fits_resource_node("hasPixelAxis"),
                         axis);
    }

    // terminate the list of axes
    assert(axis_list_p != NULL);
    librdf_model_add(model, axis_list_p, RDF_NODE(rest), RDF_NODE(nil));

    // and add the axis-list to the WCS
    assert(axis_list_head != NULL);
    librdf_model_add(model,
                     wcs_node,
                     fits_resource_node("hasPixelAxisList"),
                     axis_list_head);

    return 0;
}

int add_world_axes(fitsfile* fptr, librdf_model* model, librdf_node* wcs_node, int naxis)
{
    librdf_node* axis_list_head = NULL;
    librdf_node* axis_list_p = NULL;
    int axisnumber;
    int status = 0;
    //double dvalue;

    assert(naxis >= 1);

    for (axisnumber=1; axisnumber<=naxis; axisnumber++) {
        librdf_node* axis = librdf_new_node_from_uri_string(world, UFORMAT(0, "#wa%d", axisnumber));
        librdf_node* previous = axis_list_p;
        char string_val[80];

        axis_list_p = get_new_list_node(model, previous, axis);
        if (axis_list_head == NULL) {
            axis_list_head = axis_list_p;
        }

        librdf_model_add(model, axis, RDF_TYPE, fits_resource_node("WorldAxis"));

        librdf_model_add_typed_literal_statement(model,
                                                 axis,
                                                 fits_resource_node("axisNumber"),
                                                 UFORMAT(0, "%d", axisnumber), NULL,
                                                 xsd_predicate_uri("integer"));
        // CRVALi
        add_double_predicate(fptr, model, axis,
                             format(0, "CRVAL%d", axisnumber), "referenceCoordinate");
        add_double_predicate(fptr, model, axis,
                             format(0, "CDELT%d", axisnumber), "coordinateDelta");

        // CUNITi
        if (fits_read_key(fptr, TSTRING, format(0, "CUNIT%d", axisnumber), string_val, NULL, &status)) {
            status = 0;             // do nothing
        } else {
            librdf_model_add_string_literal_statement(model,
                                                      axis,
                                                      fits_resource_node("axisUnit"),
                                                      (unsigned char*)string_val, NULL, 0);
        }

        // CTYPEi
        if (fits_read_key(fptr, TSTRING,  format(0, "CTYPE%d", axisnumber), string_val, NULL, &status)) {
            status = 0;
        } else {
            librdf_model_add(model,
                             axis,
                             fits_resource_node("coordinateType"),
                             new_typed_node(FITS_CTYPE_BASE, string_val));
        }                     

        // PVi_m
        {
            librdf_node* head = NULL;
            librdf_node* p = NULL;
            int list_index;
            double v;
            int status = 0;
            for (list_index=1; ; list_index++) {
                if (fits_read_key(fptr, TDOUBLE, format(0, "PV%d_%d", axisnumber, list_index),
                                  &v, NULL, &status)) {
                    status = 0;
                    break;      // end of parameters -- jump out of list
                } else {
                    librdf_node* previous = p;
                    p = get_new_list_node(model,
                                             previous,
                                             librdf_new_node_from_typed_literal(world,
                                                                                UFORMAT(0, "%g", v),
                                                                                NULL, 
                                                                                xsd_predicate_uri("double")));
                    if (head == NULL) {
                        head = p;
                    }
                }
            }
            if (head != NULL) {
                // there was at least one PVi_m
                // terminate the list
                librdf_model_add(model, p, RDF_NODE(rest), RDF_NODE(nil));
                librdf_model_add(model, axis, fits_resource_node("projectionParameters"), head);
            }
        }
        // and do it again for PSi_m...
        {
            librdf_node* head = NULL;
            librdf_node* p = NULL;
            int list_index;
            int status = 0;
            char v[80];
            for (list_index=1; ; list_index++) {
                if (fits_read_key(fptr, TSTRING, format(0, "PS%d_%d", axisnumber, list_index),
                                  &v, NULL, &status)) {
                    status = 0;
                    break;      // end of parameters -- jump out of list
                } else {
                    librdf_node* previous = p;
                    p = get_new_list_node(model,
                                             previous,
                                             librdf_new_node_from_literal(world,
                                                                          (unsigned char*)v,
                                                                          NULL, 0));
                    if (head == NULL)
                        head = p;
                }
            }
            if (head != NULL) {
                // there was at least one PSi_m
                // terminate the list
                librdf_model_add(model, p, RDF_NODE(rest), RDF_NODE(nil));
                librdf_model_add(model, axis, fits_resource_node("projectionParameterStrings"), head);
            }
        }

        // Finally, add a direct relation between the WCS node and this node,
        // supplementary to the eventual inclusion in the hasWorldAxisList list.
        librdf_model_add(model,
                         wcs_node,
                         fits_resource_node("hasWorldAxis"),
                         axis);
    }

    // terminate the list of axes
    assert(axis_list_p != NULL);
    librdf_model_add(model, axis_list_p, RDF_NODE(rest), RDF_NODE(nil));

    // and add the axis-list to the WCS
    assert(axis_list_head != NULL);
    librdf_model_add(model,
                     wcs_node,
                     fits_resource_node("hasWorldAxisList"),
                     axis_list_head);

    return 0;
}

// Convenience function for adding a double-valued predicate
int add_double_predicate(fitsfile* fptr,
                         librdf_model* model,
                         librdf_node* subject, // the subject of the predicate
                         char* fits_keyword, // the FITS keyword to provide the value
                         char* rdf_predicate) // the string name of the predicate in the FITS ontology
{
    int status = 0;
    double dvalue;

    if (fits_read_key(fptr, TDOUBLE, fits_keyword, &dvalue, NULL, &status)) {
        status = 0;             // not there -- do nothing
    } else {
        librdf_model_add_typed_literal_statement(model,
                                                 subject,
                                                 fits_resource_node(rdf_predicate),
                                                 UFORMAT(0, "%g", dvalue), NULL,
                                                 xsd_predicate_uri("double"));
    }
    return 0;
}

// Convenience function.  Given an array of values, such as PCi_j,
// find all of the keywords which have headers in them, and format them
// into a list of values in a string.
int add_array_predicate(fitsfile* fptr,
                        librdf_model* model,
                        librdf_node* wcs_node,
                        char* fits_prefix, // eg, PC or CD
                        char* rdf_property, // eg, pcMatrix
                        int naxis)
{
    double dvalue;
    double *darray;
    int found_one, i, j;
    int status = status;
        
    MALLOC(darray, double, (naxis*naxis));
    found_one = 0;
    for (i=0; i<naxis; i++) {
        for (j=0; j<naxis; j++) {
            if (fits_read_key(fptr, TDOUBLE,
                              format(0, "%s%d_%d", fits_prefix, i+1, j+1),
                              &dvalue, NULL, &status)) {
                darray[i*naxis+j] = (i==j ? 1.0 : 0.0);
                status = 0;
            } else {
                darray[i*naxis+j] = dvalue;
                found_one = 1;
            }
        }
    }
    if (found_one) {
        librdf_model_add_typed_literal_statement(model,
                                                 wcs_node,
                                                 fits_resource_node(rdf_property),
                                                 darray_to_string(darray, naxis*naxis), NULL,
                                                 xsd_predicate_uri("string"));
    }
    free(darray);

    return 0;
}

double get_value_double(fitsfile* fptr, char* keyname)
{
    double v;
    int status = 0;
    if (fits_read_key(fptr, TDOUBLE, keyname, &v, NULL, &status)) {
        fits_report_error(stderr, status);
        return 0.0;
    } else {
        return v;
    }
}

int get_value_int(fitsfile* fptr, char* keyname)
{
    int v;
    int status = 0;
    if (fits_read_key(fptr, TINT, keyname, &v, NULL, &status)) {
        fprintf(stderr, "get_value_int: error retrieving key %s\n", keyname);
        fits_report_error(stderr, status);
        return 0;
    } else {
        return v;
    }
}
    
void fitserror(int status)
{
    if (status) {
        fits_report_error(stderr, status);
        exit(status);
    }
    return;
}

void general_error(char* fmt, ...)
{
    va_list ap;
    fputs("FITS error: ", stderr);
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    fputc('\n', stderr);
    exit(1);
}

// librdferror: pretty brutal just now...
void librdferror(char* msg)
{
    fputs(msg, stderr);
    fputc('\n', stderr);
    exit(1);
}

int serialize_model(librdf_model* model, librdf_uri* base, char* syntax, FILE* f)
{
    librdf_serializer* serializer;
    int status;
    
    if ((serializer = librdf_new_serializer(world, syntax, NULL, NULL)) == NULL) {
        librdferror("Can't create turtle serializer");
    }
    librdf_serializer_set_namespace(serializer,
                                    librdf_new_uri(world, (unsigned char*)FITS_CORE_BASE),
                                    "fits");
    librdf_serializer_set_namespace(serializer,
                                    librdf_new_uri(world, (unsigned char*)FITS_CTYPE_BASE),
                                    "ctype");
    librdf_serializer_set_namespace(serializer,
                                    librdf_new_uri(world, (unsigned char*)PROVENANCE),
                                    "prov");
    status = librdf_serializer_serialize_model_to_file_handle(serializer, f, base, model);
    librdf_free_serializer(serializer);
    return status;
}

// Given an existing pointer to data, p, a required_len and a pointer
// to the current length of the data, make sure that p points to a
// space as least as big as required_len, realloc'ing as necessary.
// Return a possibly updated p and *current_len.  If *current_len is
// zero, then p is ignored and new space is allocated.
char* allocate_at_least(char* p, int required_chars, size_t *current_len)
{
    if (*current_len < required_chars) {
        if (*current_len == 0) {
            *current_len = required_chars;
            MALLOC(p, char, required_chars);
        } else {
            *current_len = required_chars;
            REALLOC(p, char, required_chars);
        }
    }
    return p;
}

// fits_resource_node : string -> librdf_node
// Return a new node in the FITS namespace, including the given fragment
librdf_node* fits_resource_node(char* fragment)
{
    return new_typed_node(FITS_CORE_BASE, fragment);
}

librdf_uri* xsd_predicate_uri(char* fragment)
{
    return new_typed_uri("http://www.w3.org/2001/XMLSchema#", fragment);
}

librdf_node* new_typed_node(char* base, char* fragment)
{
    static char* buf = NULL;
    static size_t buflen = 0;

    buf = allocate_at_least(buf, strlen(base) + strlen(fragment) + 1, &buflen);
    snprintf(buf, buflen, "%s%s", base, fragment);
    return librdf_new_node_from_uri_string(world, (unsigned char*)buf);
}

librdf_uri* new_typed_uri(char* base, char* fragment)
{
    static char* buf = NULL;
    static size_t buflen = 0;

    buf = allocate_at_least(buf, strlen(base) + strlen(fragment) + 1, &buflen);
    snprintf(buf, buflen, "%s%s", base, fragment);
    return librdf_new_uri(world, (unsigned char*)buf);
}

/**
 * new_internal_node : -> librdf_node
 * Create a new blank node.  We add our own blank-identifier purely so that the identifiers are tidier!
 */
librdf_node* new_internal_node()
{
    static int blank_node_counter = 0;
    librdf_node* new_node;
#define BUFLEN 8
    char buf[BUFLEN];

    if (snprintf(buf, BUFLEN, "n%d", blank_node_counter) >= BUFLEN) {
        fprintf(stderr, "Ooops: new_internal_node ran out of integers!\n");
        exit(1);
    }
    blank_node_counter++;
    if ((new_node = librdf_new_node_from_blank_identifier(world, (unsigned char*)buf)) == NULL) {
        librdferror("Couldn't create new blank node");
    }
    return new_node;
#undef BUFLEN
}

// Return a single new rdf:List node whose rdf:first is first.  If
// previous is not NULL, then this node is tacked on to the end of
// that one (by making it that node's rdf:rest).  We do not terminate
// the list (because we can't subsequently overwrite this, so we'd end
// up with a list with two rdf:rest predicates).
librdf_node* get_new_list_node(librdf_model* model, librdf_node* previous, librdf_node* first)
{
    librdf_node* new_node = new_internal_node();
    librdf_model_add(model, new_node, RDF_NODE(first), first);
    librdf_model_add(model, new_node, RDF_TYPE, RDF_NODE(List));
    if (previous != NULL) {
        librdf_model_add(model, previous, RDF_NODE(rest), new_node);
    }
    return new_node;
}

#if 0
// Return a newly-allocated rdf:List of nodes.
// The return value is a pointer to the node at the head of the list.
// On return, *node_list_ptr is set to an array of librdf_node*; this should be freed after use.
librdf_node* get_node_list(librdf_model* model, librdf_node*** node_list_ptr, int n, char* nodenames)
{
    librdf_node** buf;
    librdf_node* head = NULL;
    librdf_node* p = NULL;
    int i;

    MALLOC(buf, librdf_node*, n);
    for (i=0; i<n; i++) {
        librdf_node* previous = p;
        if (nodenames == NULL)
            buf[i] = new_internal_node();
        else
            buf[i] = librdf_new_node_from_uri_string(world, UFORMAT(0, "#%s%d", nodenames, i+1));
        p = get_new_list_node(model, previous, buf[i]);
        if (head == NULL)
            head = p;
    }
    // terminate the list
    librdf_model_add(model, p, RDF_NODE(rest), RDF_NODE(nil));
    *node_list_ptr = buf;
    return head;
}
#endif

// format : int string ... -> string
// Format information into a string, which we return.
// If 'allocate' is non-zero, then we allocate a new string.
// If 'allocate' is zero, the string is allocated statically, so should be used immediately, and not freed.
char* format(int allocate, char* fmt, ...) 
{
    static char* buf;
    static size_t buflen = 0;
    int required_chars = strlen(fmt) + 1; /* estimate -- needn't be accurate (the +1 forces it non-zero) */
    int allocate_and_print = 1;

    while (allocate_and_print) {
        va_list ap;
        size_t size_actually_required;
        
        buf = allocate_at_least(buf, required_chars, &buflen);
        va_start(ap, fmt);
        size_actually_required = vsnprintf(buf, buflen, fmt, ap);
        va_end(ap);
        if (size_actually_required >= buflen) {
            // buffer was too small --
            // Note that the vsnprintf spec doesn't say whether size_actually_required is accurate,
            // or can only be compared with >=.  Assume the latter and ignore the actual value.
            required_chars *= 2;
        } else {
            // all OK
            allocate_and_print = 0;
        }
    }

    if (allocate) {
        char* newbuf;
        int newbuflen = strlen(buf)+1;
        MALLOC(newbuf, char, newbuflen);
        memcpy(newbuf, buf, newbuflen);
        return newbuf;
    } else {
        return buf;
    }
}

// darray_to_string : double* int -> char*
// Given an array, return a string which has these formatted into it
unsigned char* darray_to_string(double* d, int n)
{
    static char *buf;
    // perhaps need to think about the precision here
#define PRECISION 8
    static size_t buflen = 0;
    int required = n * (6 + PRECISION + 1) + 1; /* -d.ddd...edd */
    int i, bufi;
    
    buf = allocate_at_least(buf, required, &buflen);
    for (bufi=0, i=0; i<n; i++) {
        int nprinted;
        nprinted = snprintf(&buf[bufi], buflen-bufi, "%.*e ", PRECISION, d[i]);
        bufi += nprinted;
    }
    return (unsigned char*)buf;
}

void Usage()
{
    fprintf(stderr, "Usage: %s [options] filename [base-uri]\n", progname);
    fprintf(stderr, "\nOptions:\n  -s fmt : serialization format (turtle/rdfxml/ntriples)\n");
    exit(1);
}
