#lang racket

(require (prefix-in rdf: rdf/rdf)
         scheme/cmdline)

(struct fits-card (key value))

(define (fits-uri fragment)
  (rdf:new-uri (string-append "http://purl.org/astro/fits/core#" fragment)))
(define ctype-base "http://purl.org/astro/fits/ctype#")

;; Each of these functions is wcs-node -> (listof fits-card)
(define (general-header-information wcs model)
  (list (fits-card "SIMPLE" #t)
        (let ((hdu (rdf:find-node/node model #f 'a (fits-uri "FITSheader"))))
          (cond ((not hdu)
                 (error "Ooops: I can't seem to find a FITSheader object!"))
                ((rdf:find-node model hdu (fits-uri "bitsPerPixel") #f)
                 => (λ (bitpix)
                       (fits-card "BITPIX" (rdf:literal->value bitpix))))
                (else (error "No bitsPerPixel property"))))))

(define (pixel-axis-information wcs model)
  (let ((pixel-axes (rdf:find-node model wcs (fits-uri "hasPixelAxisList") #f)))
    (list (fits-card "NAXIS" (length pixel-axes))
          (if pixel-axes
              (map (λ (axis)
                      (define (get s) (rdf:find-node model axis (fits-uri s) #f))
                      (let ((axisn (rdf:literal->value (get "axisNumber"))))
                        (list (cond ((get "referencePixel") ;(rdf:find-node model axis (fits-uri "referencePixel") #f)
                                     => (λ (crpix)
                                           (fits-card (format "CRPIX~a" axisn)
                                                      (rdf:literal->value crpix))))
                                    (else (error "No referencePixel property!")))
                              (cond ((get "numberOfPixels")
                                     => (λ (naxis)
                                           (fits-card (format "NAXIS~a" axisn)
                                                      (rdf:literal->value naxis))))
                                    (else (error "no numberOfPixels property!"))))))
                   pixel-axes)
              (error "No hasPixelAxisList property!")))))
#;(define (xx-pixel-axis-information wcs model)
  (let ((pixel-axes (rdf:fold-nodes cons '() model wcs (fits-uri "hasPixelAxisList") #f)))
    (unless (= (length pixel-axes) 1)
            (error (format "Wrong number of hasPixelAxisList properties: ~s~%" pixel-axes)))
    (apply append (map (λ (axis)
                          (let ((axisn (rdf:find-node model axis (fits-uri "axisNumber") #f)))
                            (cond ((rdf:find-node model axis (fits-uri "referencePixel") #f)
                                   => (λ (crpix)
                                         (list (fits-card (format "CRPIX~a" (rdf:literal->value axisn))
                                                          (rdf:literal->value crpix)))))
                                  (else (error "No referencePixel property!")))))
                       (car pixel-axes)))))

(define (world-axis-information wcs model)
  (map (lambda (axis)
         (define (get s) (rdf:find-node model axis (fits-uri s) #f))
         (let ((axisn (rdf:literal->value (get "axisNumber"))))
           (list (cond ((get "referenceCoordinate")
                        => (lambda (coord)
                             (fits-card (format "CRVAL~a" axisn)
                                        (rdf:literal->value coord))))
                       (else #f))
                 (fits-card 'comment (format "axis number ~a" axisn))
                 (cond ((get "projectionParameters")
                        => (lambda (paramlist)
                             (let loop ((l paramlist)
                                        (n 1))
                               (if (null? l)
                                   '()
                                   (cons (fits-card (format "PV~a_~a" axisn n)
                                                    (rdf:literal->value (car l)))
                                         (loop (cdr l) (+ n 1)))))))
                       (else #f))
                 (cond ((get "projectionParameterStrings")
                        => (lambda (paramlist)
                             (let loop ((l paramlist)
                                        (n 1))
                               (if (null? l)
                                   '()
                                   (cons (fits-card (format "PS~a_~a" axisn n)
                                                    (rdf:literal->value (car l)))
                                         (loop (cdr l) (+ n 1)))))))
                       (else #f))
                 (cond ((get "coordinateType")
                        => (lambda (ctype-uri)
                             (cond ((regexp-match (regexp (string-append ctype-base "(.+)")) (rdf:thing->string ctype-uri))
                                    => (λ (ctype-list)
                                          (fits-card (format "CTYPE~a" axisn)
                                                     (cadr ctype-list))))
                                   (else (fits-card 'comment (format "Unexpected coordinate type ~a" ctype-uri))))))
                       (else #f))
                 ;; how do I format the coordinateType URL?
                 )))
       (or (rdf:find-node model wcs (fits-uri "hasWorldAxisList") #f) '())))

(let ((input-file (command-line #:program "rdf2fits"
                                #:args (rdf-file)
                                rdf-file)))
  (define (write-fits-list! l)
    (cond ((or (not l) (null? l)) #f)
          ((list? l) (for-each write-fits-list! l))
          ((boolean? (fits-card-value l))
           (printf "~a = ~a~%" (fits-card-key l) (if (fits-card-value l) "T" "F")))
          ((eqv? (fits-card-key l) 'comment)
           (printf "COMMENT   ~a~%" (fits-card-value l)))
          ((string? (fits-card-value l))
           (printf "~a = '~a'~%" (fits-card-key l) (fits-card-value l)))
          (else (printf "~a = ~a~%" (fits-card-key l) (fits-card-value l)))))
  #;(write-fits-list!
   (reverse
    (let* ((model (rdf:add-to-model #f (string->path input-file)))
           (wcs-node (rdf:find-node/node model #f 'a (fits-uri "WCSInformation"))))

      (foldl (λ (proc card-list)
                (append (proc wcs-node model) card-list))
      '()
      (list general-header-information
            pixel-axis-information
            world-axis-information)))))
  (let* ((model (rdf:add-to-model #f (string->path input-file)))
         (wcs-node (rdf:find-node/node model #f 'a (fits-uri "WCSInformation"))))

      (for-each (λ (proc)
                   (write-fits-list! (proc wcs-node model)))
                (list general-header-information
                      pixel-axis-information
                      world-axis-information))))
